/* ------------------------------------------------------------------------
	poster.c

	A Citadel utility to post files as multi-part citnet messages.

	Written by Audin Malmin.

    Poster is free software; you can redistribute it and/or modify it
    under the terms of the GNU Public License as published by the Free
    Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    Write to the Free Software Foundation, Inc., 675 Mass Ave,
    Cambridge, MA 02139, USA for a copy of the license.

	There is, of course, no warranty what so ever.  This code works fine
	on my system and the few others it's been used on.  But if it makes
	your system melt down in front of your eyes then its your own fault.

	If you have questions or comments, I can be reached at:

	Magrathea @ Skaro, Olympia [skr.olm]  (CitNet)

	amalmin@halcyon.com  (Internet)

	or, if all else fails:

	Audin Malmin
	1317 Horne St. N.E.
	Olympia, WA 98516-5613

 --------------------------------------------------------------------------
 History:  (is also in a seperate file...poster.txt.)
 10-2?-96
    cfg and msgbuf structures.
    Cleaned up the Source ID handling.
    putmsg() aborts if a required field is missing.
    putmsg() is now more generic.
    clean_msgbuf() added.
    reset() added.
    startup() handles more.
    startup() returns a value based on file problems...
    msgbuf.cmt_ptr...a linked list of comments...
    add_cmt(), remove_cmt(), clean_cmts().
    local-only messages...  two attribute bytes...
 10-24-96
    PutField().
    no more PutStr1().
    Lots of reformating.
    putmsg() is even more generic.
    startstr is completely handled in main, not in putmsg.
    "Precidence: Bulk" is handled in main, not putmsg.
 10-26-96
    process_file().
    Many little things moved around to support the new interface...
    size_cmt().
    Spaces = 2.  (Wow...Only been meaning to do THAT for maybe two years...)
    newroom = ROOM_[NO]APPEND stuff to support multiple roomnames in one
        execution...
    Rewrote the startup display to only display the helpful portions...
 10-28-96
    Spaces = 2 worked on a bit more.
    more startup display twiddling.
 10-29-96
    num_parts() finished...
    cfg.abbey_road added.  #accuracy.
    size comment function takes cfg.abbey_road into account...
    -f command line option...(formatting control).
 11-01-96
    -c command line option...(config filename).
 11-11-96
    LINESIZE, BIGLABEL defines...
    Fixed a few possible overruns in config().
    Fixed that damned fileposition bug.
    Cleaned up cfg.spaces control a bit more.
    Few extra comments.
    Config file accepts "1", "Yes", or "On" as positive arguments.
    GNU public license stuff.
 -------------------------------------------------------------------------- */
#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>
#include  <ctype.h>
#include  <conio.h>
#include  <dos.h>
#include  <dir.h>
#include  <time.h>
#include  <io.h>
#include  "..\stat\pragmas.h"   // merely turns on a bunch of compiler warnings.

#define   TRUE          1
#define   FALSE         0
#define   ROOM_NOAPPEND 1
#define   ROOM_APPEND   2

#define	  LABELSIZE     30
#define   BIGLABEL      80
#define   LINESIZE      256

typedef char label[LABELSIZE + 1];

char    version[] = "Poster 3.00";

typedef struct comment {
        char text[LINESIZE + 1];
        struct comment *next;
} comment;

comment *currcmt = NULL;

struct {
    FILE    *in, *out, *altout;
    label   room;
    int     spaces,             // 0 = ^A5, 1 = ' ' (spaces)
            unix,               // 1 = Strip unix mail header.
            partsize,           // Part size.  Default = 8000.
            append,             // allow appending to an existing packet.
            montyl,             // Output the message text to another file
            local,              // Post local-only messages?
            abbey_road;         // use accurate part number stuff...
    long    startpoint,         // # of lines to skip at start of file.
            minfsize,           //
            maxfsize;           //
    char    input[LINESIZE + 1],         // Input file.
            configfile[LINESIZE + 1],
            altname[LINESIZE + 1],
            progname[LINESIZE + 1],      // Used in the help menu.
            shrt_pname[LABELSIZE + 1],     // just the "poster.exe" part.
            startstr[LINESIZE + 1],      // The string to START posting from.
            endstr[LINESIZE + 1];        // The string to STOP posting on.
} cfg;

struct {     /* Message buffer. */
    int      ID;                    // Local ID field.  Ignored by Citadel.
    label    SID,                   // S - Source ID
             author,                // A - Author
             realname,              // ! - Authors real name
             date,                  // D - datestamp
             title,                 // N - Title
             surname,               // n - Surname
             room,                  // R - Room name
             group,                 // G - Group
             reply_msg,             // I - Message # being replied to
             mail_to,               // T - mail recipient
             node_to,               // Z - Destination node
             region_to,             // z - Destination node's city
             state_to,              // q - Destination node's state
             country_to,            // u - Destination node's country
             phone_to,              // h - Destination node's phone number
             address_to,            // > - destination node's address
             public_to,             // f - Public-to field.
             node,                  // O - Originating node
             address,               // P - Node address first part of P>ath.
             state,                 // Q - Originating node's state
             region,                // o - Originating node's region/city.
             country,               // U - Originating node's country code
             phone,                 // H - Originating node's Phone number
             twit_region,           // J - Originating node's Twit region
             twit_country,          // j - Originating node's Twit country
             software,              // s - Originating node's software
             stamp;                 // % - Incorporation Time Stamp
    char     attr,                  // First attribute byte.
             attr2,                 // Second attribute byte.
             text[8193],            // M - The main body of the message.
             subject[BIGLABEL + 1], // B - Subject of the messages...
             sig1[BIGLABEL + 1],    // _ - User Signature line.
             sig2[BIGLABEL + 1],    // . - BBS Signature line.
             fpath[LINESIZE + 1],   // p - The Forced Path field.
             gateway[LINESIZE + 1]; // * - The '*' gateway field.
    comment  *cmt_ptr;              // Pointer to first comment.
} msgbuf;

time_t	timestamp;
long	SID;
int		numcmts      = 0;           // total number of comments.
int     newroom      = ROOM_NOAPPEND;
int     multi_room   = -1;
int     multi_file   = 0;
int     multi_author = 0;

int     startup(void);                         // initialize everything
int		config(void);                          // read the config file.
void	envr(void);                            // read the environment.
int     process_file(char *input);             // post one file.
void    num_msgs(long *filesize, int *totalparts); // figures number of parts, etc
void    size_cmt(int totalparts, char *input_name, long filesize, int part);
int     open_file(char *input, long *filesize, int *totalparts);
int     putmsg(FILE *out);                     // Output a message
int 	GetBlock(char *str, int mlen);         // Get the file text.
void 	PutStr(FILE *fl, char *str);           // Put a Z-string to fl.
void    PutField(FILE *fl, char field, char *text); // put out a msg field.
void    clean_msgbuf(void);                    // Clear out the message buf.
int     add_cmt(char *str);                    // Add a comment to the msgbuf.
int     remove_cmt(void);                      // removes last comment.
int     clean_cmts(void);                      // remove all comments.
void    reset(void);                           // resets all the config stuff.
void	usage(void);                           // Show the usage screen.

/* Main */
int main(int argc, char *argv[])
{
int exit_val = 0;

	strcpy(cfg.progname, argv[0]);
	strlwr(cfg.progname);

    fnsplit(cfg.progname, NULL, NULL, cfg.shrt_pname, NULL);

    argc--;
    argv++;

	printf("\n"
		   "%s (%s %s)\n"
		   "\n"
		   "By Magrathea @ Skaro, Olympia [skr.olm]\n"
		   "amalmin@halcyon.com\n"
		   "\n", version, __TIME__, __DATE__);

	if(argc < 1) {
		usage();
        return(0);
    }

    reset();        // reset all config options.
    clean_msgbuf(); // Null out the message buffer.

    if(toupper(*(*argv + 1)) == 'C') {
        argv++;                         // Handles an alternate starting
        argc--;                         // config file.
        strcpy(cfg.configfile, *argv);
        argv++;
        argc--;
    }

	config();       // Read the config file.
	envr();         // Read the Environment.

    if((exit_val = startup()) != 0) {   // Opens output files...
        if(exit_val != 69)
            unlink("msg.69");
        printf("Exiting with errorlevel = %d\n", exit_val);
        return(exit_val);
    }

    while(argc > 0) {
        if(**argv == '-') {
            switch(toupper(*(*argv + 1))) {
                case 'R':                   // New room.
                    argv++;
                    argc--;
                    if(*argv == 0 || stricmp(cfg.room, *argv) == NULL)
                        break;
                    strcpy(cfg.room, *argv);
                    printf("Roomname: %s\n", cfg.room);
                    if(multi_room < 0)
                        newroom = ROOM_NOAPPEND;
                    else
                        newroom = ROOM_APPEND;
                    break;

                case 'A':                   // New Author.
                    argv++;
                    argc--;
                    if(*argv) {
                        strcpy(msgbuf.author, *argv);
                        printf("Author: %s\n", msgbuf.author);
                    }
                    multi_author++;
                    break;

                case 'C':                   // New config file.
                    argv++;
                    argc--;
                    if(*argv) {
                        printf("Config file: %s\n", *argv);
                        fcloseall();
                        puts("Close all files.");
                        reset();
                        puts("Reset config options.");
                        clean_msgbuf();
                        puts("Reset message buffer.");
                        strcpy(cfg.configfile, *argv);
                        config();
                        puts("Read new config file.");
                        puts("Open everything back up.");
                        if((exit_val = startup()) != 0) {
                            if(exit_val != 69)           // open output files
                                unlink("msg.69");
                            printf("Exiting with errorlevel = %d\n", exit_val);
                            return(exit_val);
                        }
                    }
                    break;

                case 'L':                   // Local/Network messages...
                    if((*(*argv + 2)) == '+') {
                        cfg.local = 1;
                        puts("Local messages.");
                    } else {
                        cfg.local = 0;
                        puts("Networked messages.");
                    }
                    break;

                case 'F':                   // New formatting method.
                    switch(*((*argv) + 2))
                        {
                        case '0':
                            cfg.spaces = 0;
                            puts("Using ^A5 for formatting.");
                            break;
                        default:
                            puts("Invalid -f parameter");
                            /* fall through */
                        case '1':
                            cfg.spaces = 1;
                            puts("Using spaces for formatting.");
                            break;
                        case '2':
                            cfg.spaces = 2;
                            puts("Intelligent formatting control.");
                            break;
                        }
                    break;
                default:
                    printf("Unknown command line option: %s\n", *argv);
                    break;
            }
        } else {
            if(process_file(*argv) == FALSE)    // Post the file.
                printf("File \"%s\" was not posted.\n", *argv);
        }
        argv++;
        argc--;
    }

    printf("%d file%s %s posted to %d room%s by %d author%s.\n",
           multi_file, multi_file > 1 ? "s" : "",
           multi_file > 1 ? "were" : "was",
           multi_room + 1, multi_room + 1 > 1 ? "s" : "",
           multi_author, multi_author > 1 ? "s": "");

	return(0);
}

/* ------------------------------------------------------------------------ */
/* process_file() -- posts one file.  (filename as arg)                     */
/* ------------------------------------------------------------------------ */
int process_file(char *input)
{
char input_name[31] = { 0 };
char input_ext[5]   = { 0 };
char *endpt         = 0;
long filesize       = 0;
int  good           = 0,
     part           = 1,
     totalparts     = 0;

	fnsplit(input, NULL, NULL, input_name, input_ext);
	strcat(input_name, input_ext);

    printf("Posting \"%s\"", input_name);

    if(open_file(input, &filesize, &totalparts) == FALSE)
        return(FALSE);

    printf(", %ld bytes, %d parts.\n", filesize, totalparts);

    if(newroom) {
        if(newroom == ROOM_APPEND)
            fputc(0x00, cfg.out);

        printf("  Setting new room: %s %s\n", cfg.room, newroom == ROOM_APPEND ? "(appending)" : "");
        PutStr(cfg.out, cfg.room);
        newroom = 0;
        multi_room++;
    }

    /*
	 * file posting loop.
     */

	while(!good && !endpt)
    {
        if(!cfg.spaces) {
            strcpy(msgbuf.text, "\x01\x35\x01\x30\x01\x61"
                                "Precedence: Bulk" "\x01\x30\n");
        } else {
            strcpy(msgbuf.text, "\x01\x30\x01\x61"
                                "Precedence: Bulk" "\x01\x30\n");
            if(cfg.spaces == 1)
                strcat(msgbuf.text, " ");
/*            else
                if(cfg.spaces == 2)
                    strcat(msgbuf.text, "\t");
*/
        }

        if(part == 1 && *cfg.startstr) {
            printf("Start String is: %s\n", cfg.startstr);
            strcat(msgbuf.text, cfg.startstr);
            }

		good = GetBlock(msgbuf.text + strlen(msgbuf.text),
                      cfg.partsize - strlen(msgbuf.text));

		if(*cfg.endstr)
			if((endpt = strstr(msgbuf.text, cfg.endstr)) != NULL)
				*endpt = 0;  // Null off the message text.

		if(cfg.montyl) {     // Make a copy of the message text.
			if(part == 1 && *cfg.startstr) {
				fputs(cfg.startstr, cfg.altout);
			}
			fputs(msgbuf.text, cfg.altout);
		}

        ltoa(SID, msgbuf.SID, 10);

        printf("    writing message # %d of %d (SID = %s) %d bytes\n",
                part, totalparts, msgbuf.SID, (int)strlen(msgbuf.text));

        size_cmt(totalparts, input_name, filesize, part);

        if(!putmsg(cfg.out)) {    // output the message...
            printf("Exiting with errorlevel = 3\n");
            exit(3);
        }
        remove_cmt();       // Remove the now old size comment.

        SID++;              // generate a new source ID.
        part++;             // New part number.
		*msgbuf.text = 0;
	}                      /* end main loop */

    putc('\n', stdout);

    if(*cfg.endstr && endpt == NULL)
        printf("  Never found ending string in file %s, whole file posted.\n", input);

    fclose(cfg.in);
    multi_file++;

    return(TRUE);
}

/* ------------------------------------------------------------------------ */
/* open_file() -- opens an input file and goes through the start* stuff     */
/* ------------------------------------------------------------------------ */
int open_file(char *input, long *filesize, int *totalparts)
{
int  done = 0, i;

	if ((cfg.in = fopen(input, "rb")) == NULL) {
		printf("  I can't seem to find \"%s\".\n", input);
		return(FALSE);
	}

	if((*filesize = filelength(fileno(cfg.in))) == 0) {
		printf("  Duh...Your input file \"%s\" is empty...\n", input);
        fclose(cfg.in);
		return(FALSE);
	}

    if(cfg.minfsize && *filesize < cfg.minfsize) {
        printf("  Your input file \"%s\" is smaller than your MINSIZE setting.\n", input);
        fclose(cfg.in);
        return(FALSE);
    }

    if(cfg.maxfsize && *filesize > cfg.maxfsize) {
        printf("  Your input file \"%s\" is larger than your MAXSIZE setting.\n", input);
        fclose(cfg.in);
        return(FALSE);
    }

	if(cfg.startpoint != 0) {
		for(i = 0; i<cfg.startpoint;i++) {
            if(fgets(msgbuf.text, LINESIZE, cfg.in) == NULL) {
                printf("  Never reached start point in file %s.\n", input);
                fclose(cfg.in);
                return(FALSE);
            }
        }
		msgbuf.text[0] = 0;
	}

    if(cfg.unix) {
        do {
            fgets(msgbuf.text, sizeof(msgbuf.text), cfg.in);
        } while(msgbuf.text[0] != '\n' && !feof(cfg.in));

        if(feof(cfg.in)) {
            printf("  Never found the end of the mail header.\n");
            fclose(cfg.in);
            return(FALSE);
        }
    }

	done = 0;

	if(*cfg.startstr) {
        while(!done && fgets(msgbuf.text, (int)sizeof(msgbuf.text), cfg.in) != NULL) {
            if(strstr(msgbuf.text, cfg.startstr) != NULL) {
                printf("String: %s\n", msgbuf.text);
                *cfg.startstr = 0;
                strncat(cfg.startstr, msgbuf.text, sizeof(cfg.startstr));
                done = 1;
            }
        }

        if(!done) {
            printf("  Never found start string in file %s.\n", input);
            fclose(cfg.in);
            return(FALSE);
        }
    }

    if(cfg.abbey_road)                  // Accurate parts detection.
        num_msgs(filesize, totalparts); // Figure out the number of parts.
    else
        *totalparts = (int)((*filesize / cfg.partsize) + 1);

    return(TRUE);
}

/* ------------------------------------------------------------------------ */
/* size_cmt() -- works up a size comment.                                   */
/* ------------------------------------------------------------------------ */
void size_cmt(int totalparts, char *input_name, long filesize, int part)
{
char cmttemp[LINESIZE + 1]   = { 0 };

    if(totalparts == 1) {
	    sprintf(cmttemp, "%s, %ld bytes total.",
			    input_name, filesize);
    } else {
	    if(cfg.abbey_road || (!cfg.startstr[0] && !cfg.endstr[0] && !cfg.startpoint))
		    sprintf(cmttemp, "%s, %ld bytes total, %u this part, Part %d of %d.",
				    input_name, filesize, strlen(msgbuf.text), part, totalparts);
	    else
		    sprintf(cmttemp, "%s, originally %ld bytes, %u this part, Part %d of (maybe) %d.",
				    input_name, filesize, strlen(msgbuf.text), part, totalparts);
    }

    add_cmt(cmttemp);   // add the size comment to the msgbuf.
}

/* ------------------------------------------------------------------------ */
/* size_msg() -- figures out how many parts it will take for the file.      */
/* ------------------------------------------------------------------------ */
void num_msgs(long *filesize, int *totalparts)
{
fpos_t start_ptr = 0;
char *endpt = NULL;
long bytes = 0;
int parts = 0;
int good = 0;

    if(fgetpos(cfg.in, &start_ptr))    // record the starting point.
       puts("Problem with fgetpos()");
                                    // need to post from here to stopstring.
	while(!good && !endpt) {
        if(!cfg.spaces) {
            strcpy(msgbuf.text, "\x01\x35\x01\x30\x01\x61"
                                "Precedence: Bulk" "\x01\x30\n");
        } else {
            strcpy(msgbuf.text, "\x01\x30\x01\x61"
                                "Precedence: Bulk" "\x01\x30\n");
            if(cfg.spaces == 1)
                strcat(msgbuf.text, " ");
/*            else
                if(cfg.spaces == 2)
                    strcat(msgbuf.text, "\t");
*/
        }

        if(*totalparts == 1 && *cfg.startstr)
            strcat(msgbuf.text, cfg.startstr);

		good = GetBlock(msgbuf.text + strlen(msgbuf.text),
                      cfg.partsize - strlen(msgbuf.text));

		if(*cfg.endstr)
			if((endpt = strstr(msgbuf.text, cfg.endstr)) != NULL)
				*endpt = 0;  // Null off the message text.

        parts++;
        bytes += strlen(msgbuf.text);
    }

    *totalparts  = parts;
    *filesize    = bytes;
    *msgbuf.text = 0;

    fsetpos(cfg.in, &start_ptr);
}

/* ------------------------------------------------------------------------ */
/*  putmsg() -- Writes a message out to the 'out' stream.                   */
/* ------------------------------------------------------------------------ */
int putmsg(FILE *out)
{
comment *cmt_scratch = NULL;
static char required[] = "    putmsg(): A %s field is required.\n";

	fputc(0xFF, out);              /* 0xFF to start things off... */

    fputc(msgbuf.attr, out);       // The first attribute byte...

    if((msgbuf.attr & 0x40) != 0)  // And the second attr byte...if there is
        fputc(msgbuf.attr2, out);  // one...

	/* put local ID # out */
	PutStr(out, msgbuf.SID);       // Put out the Local ID.

	/* Write out the message header... */
    if(*msgbuf.author)
        { PutField(out, 'A', msgbuf.author); }
    else
        { printf(required, "Author"); return(FALSE); }

	time(&timestamp);                  // Generate timestamp.
    ltoa(timestamp, msgbuf.date, 10);
    PutField(out, 'D', msgbuf.date);   // Output the timestamp.

    if(*msgbuf.node)
        { PutField(out, 'O', msgbuf.node); }
    else
        { printf(required, "Node"); return(FALSE); }

    if(*msgbuf.SID)
        { PutField(out, 'S', msgbuf.SID); }
    else
        { printf(required, "Source ID"); return(FALSE); }

    if(*msgbuf.address)
        { PutField(out, 'P', msgbuf.address); }
    else
        { printf(required, "Address"); return(FALSE); }

    // Output the comments...
    for(cmt_scratch = msgbuf.cmt_ptr; cmt_scratch != NULL;
         cmt_scratch = cmt_scratch->next) {
        PutField(out, '%', cmt_scratch->text);
    }

    if(*msgbuf.software)     { PutField(out, 's', msgbuf.software);     }
    if(*msgbuf.region)       { PutField(out, 'o', msgbuf.region);       }
    if(*msgbuf.state)        { PutField(out, 'Q', msgbuf.state);        }
    if(*msgbuf.country)      { PutField(out, 'U', msgbuf.country);      }
    if(*msgbuf.subject)      { PutField(out, 'B', msgbuf.subject);     	}
    if(*msgbuf.title)        { PutField(out, 'N', msgbuf.title);       	}
    if(*msgbuf.surname)      { PutField(out, 'n', msgbuf.surname);     	}
    if(*msgbuf.group)        { PutField(out, 'G', msgbuf.group);       	}
    if(*msgbuf.phone)        { PutField(out, 'H', msgbuf.phone);       	}
    if(*msgbuf.twit_region)  { PutField(out, 'J', msgbuf.twit_region); 	}
    if(*msgbuf.twit_country) { PutField(out, 'j', msgbuf.twit_country);	}
    if(*msgbuf.mail_to)      { PutField(out, 'T', msgbuf.mail_to);     	}
    if(*msgbuf.node_to)      { PutField(out, 'Z', msgbuf.node_to);     	}
    if(*msgbuf.address_to)   { PutField(out, '>', msgbuf.address_to);  	}
    if(*msgbuf.public_to)    { PutField(out, 't', msgbuf.public_to);  	}
    if(*msgbuf.phone_to)     { PutField(out, 'h', msgbuf.phone_to);     }
    if(*msgbuf.country_to)   { PutField(out, 'u', msgbuf.country_to);  	}
    if(*msgbuf.state_to)     { PutField(out, 'q', msgbuf.state_to);   	}
    if(*msgbuf.region_to)    { PutField(out, 'z', msgbuf.region_to);  	}
    if(*msgbuf.realname)     { PutField(out, '!', msgbuf.realname);   	}
    if(*msgbuf.fpath)        { PutField(out, 'p', msgbuf.fpath);        }
    if(*msgbuf.sig1)         { PutField(out, '_', msgbuf.sig1);        	}
    if(*msgbuf.sig2)         { PutField(out, '.', msgbuf.sig2);        	}
    if(*msgbuf.gateway)      { PutField(out, '*', msgbuf.gateway);      }
    if(*msgbuf.reply_msg)    { PutField(out, 'I', msgbuf.reply_msg);    }

    if(*msgbuf.text)
        PutField(out, 'M', msgbuf.text);
    else {
        printf(required, "Text");
        return(FALSE);
    }
    return(TRUE);
}

/* -------------------------------------------------------------------- */
/*  GetBlock() - Gets an mlen length chunk of text from the text file.  */
/*			   This also ads spaces as needed...(as set by spaces.)     */
/* -------------------------------------------------------------------- */
int GetBlock(char *str, int mlen)
{
char line[LINESIZE + 1] = { 0 };
int	l = 0,
    newline = 0;

    switch(cfg.spaces)              // No spaces needed, using ^A5
        {
        case 0:
            for(l = 0; l < mlen - LINESIZE + 1; l += strlen(line))
            {
                *line = 0;
                fgets(line, LINESIZE, cfg.in);
                strcat(str, line);

                if(feof(cfg.in))
                    break;
            }
            break;

        default:
            puts("Bad #formatting setting.  Using #formatting 1.");
        case 1:                     // Spaces needed.  Using compatable mode.
            for(l = 0; l < mlen - LINESIZE - 1; l += strlen(line))
            {
                line[0] = ' ';
                line[1] = '\0';
                fgets(line + 1, LINESIZE - 1, cfg.in);
                strcat(str, line);

                if(feof(cfg.in))
                    break;
            }
            break;

        case 2:
            for(l = 0; l < mlen - LINESIZE; l += strlen(line)) {
                char *ptr;
                *line = 0;
                fgets(line, LINESIZE, cfg.in);

                for(ptr = line; *ptr == ' ' || *ptr == '\t'; ptr++)
                    ;   // move through all the whitespace.

                if(*ptr == '\n') {
                    newline = 1;    // An empty line...   signals end of
                } else              // paragraph...(maybe)
                    if(newline && line[0] != ' ' && line[0] != '\t') {
                        strcat(str, "\t");   // New paragraph.
                        newline = 0;         // add a tab if not there already
                    }

                if(*line == '\n') {   // Citadel will collapse empty lines...
                    line[0] = ' ';    // so add a space so cit will leave it
                    line[1] = '\n';   // alone.
                    line[2] = 0;
                }
                strcat(str, line);
                if(feof(cfg.in))
                    break;
            }
            break;
        }

	if(!feof(cfg.in))
		return FALSE;
	else
		return TRUE;
}

/* -------------------------------------------------------------------- */
/*  PutStr()        puts a null-terminated string to a file             */
/* -------------------------------------------------------------------- */
/*					This is from DragCit.								*/
/* -------------------------------------------------------------------- */
void PutStr(FILE *fl, char *str)
{
	fwrite(str, sizeof(char), (strlen(str) + 1), fl);
}

/* -------------------------------------------------------------------- */
/*  PutField()        puts a Citadel message field to fl.               */
/* -------------------------------------------------------------------- */
void    PutField(FILE *fl, char field, char *text)
{
    fputc(field, fl);
	fwrite(text, sizeof(char), (strlen(text) + 1), fl);
}

/* -------------------------------------------------------------------- */
/*  startup()  --  startup...                                           */
/* -------------------------------------------------------------------- */
int startup(void)
{
	if(!cfg.append)	{
		if((cfg.out = fopen("msg.69", "r")) != NULL) {
            puts("Append mode needs to be on if you wish to add to an existing msg.69.\n");
            return(69);
			}

		fclose(cfg.out);

		if((cfg.out = fopen("msg.69", "wb")) == NULL) {
			printf("I need to write to \"MSG.69\", but I can't seem to open it.\n");
			return(1);
		}
	} else {
		if((cfg.out = fopen("msg.69", "r")) != NULL)
			cfg.append = 2;

        fclose(cfg.out);

		if((cfg.out = fopen("msg.69", "ab")) == NULL) {
			puts("Well, you're in append mode...but i still can't seem to open MSG.69...\n");
			return(1);
		}
	}

	if(cfg.montyl) {
		if((cfg.altout = fopen(cfg.altname, "wb")) == NULL) {
			printf("I can't seem to open %s as the Comp output file.\n", cfg.altname);
			return(1);
		}
	}

	if(cfg.append != 2) {
		PutStr(cfg.out, msgbuf.address);		// Output the address.
        newroom = ROOM_NOAPPEND;
	} else {                                    // but only if not appending.
        multi_room++;
        newroom = ROOM_APPEND;
    }

	printf("From        : [%s] %s [%s] @\n"
		   "            : %s, %s, %s [%s]\n"
           "Real name   : %-30.30s  Roomname : %s\n",
           msgbuf.title, msgbuf.author, msgbuf.surname, msgbuf.node,
           msgbuf.region, msgbuf.state, msgbuf.address,
           msgbuf.realname, cfg.room);

    if(*msgbuf.group || *msgbuf.phone)
        printf("Group       : %-30.30s  Phone    : %s\n",
               msgbuf.group, msgbuf.phone);

    if(*msgbuf.twit_region || *msgbuf.twit_country)
        printf("Twit_Region : %-30.30s  T_Country: %s\n",
               msgbuf.twit_region, msgbuf.twit_country);

    if(*msgbuf.mail_to || *msgbuf.node_to || *msgbuf.address_to)
        printf("Mail to     : %s @ %s [%s]\n",
               msgbuf.mail_to, msgbuf.node_to, msgbuf.address_to);

    if(*msgbuf.public_to) printf("Public to   : %s\n", msgbuf.public_to);
    if(*msgbuf.subject)   printf("Subject     : %s\n", msgbuf.subject);
    if(*msgbuf.sig1)      printf("User sig    : %s\n", msgbuf.sig1);
    if(*msgbuf.sig2)      printf("BBS sig     : %s\n", msgbuf.sig2);
    if(*msgbuf.fpath)     printf("Forced path : %s\n", msgbuf.fpath);
    printf("Part size   : %d\n", cfg.partsize);
    if(numcmts)          printf("Comments    : %d\n", numcmts);

    printf("\n%s%s%s%s for formatting control%s\n",
           (cfg.local) ? "Local messages, " : "Network messages, ",
           (cfg.unix) ? "Skipping unix mail header, " : "",
           (cfg.append) ? ((cfg.append == 2) ? "Appending to old MSG.69 file, " : "Append mode on, but no old MSG.69, ") : "",
           (cfg.spaces) ? ((cfg.spaces == 1) ? "Spaces" : "Intelligent spaces") : "^A5",
           (cfg.abbey_road) ? ", Accurate total parts detection." : ".");

    if(cfg.startpoint != 0) printf("Starting Point: %ld lines\n", cfg.startpoint);
    if(*cfg.startstr) printf("Starting text: \"%s\"\n", cfg.startstr);
    if(*cfg.endstr)	printf("Ending text: \"%s\"", cfg.endstr);
    if(cfg.montyl) printf("Also copying message text to: %s\n", cfg.altname);

    putc('\n', stdout);

    SID = time(&timestamp);

    strcpy(msgbuf.room, cfg.room);
    strcpy(msgbuf.software, version);

    return(0);                  // startup successful.
}

/* ------------------------------------------------------------------- */
/* config()  --  Reads a config file.                                  */
/* ------------------------------------------------------------------- */
int config(void)
{
	int i;
	FILE *cfgfile;
	char line[LINESIZE + 1], word1[LINESIZE + 1], lword1[LINESIZE + 1];
	char *text;
	char drive[5] = { 0 }, dir[LINESIZE + 1] = { 0 };

	static char *KEYWORD[] = {
		"AUTHOR",				/* 0  - 'A'uthor field            */
		"NODE",                 /* 1  - 'O'riginating Node        */
		"ADDRESS",      		/* 2  - first entry in 'P'ath     */
		"REGION",	        	/* 3  - 'o' region field          */
		"DEFAULT_ROOM",         /* 4  - 'R' field                 */
		"UNIX_MAIL",            /* 5  - start after first blank line */
		"DELAYED_START",        /* 6  - line number to start on   */
		"START_STRING",         /* 7  - The string to start with  */
		"SUBJECT",              /* 8  - 'B' Subject field         */
		"END_STRING",           /* 9  - The string to end with    */
		"COMMENT",              /* 10 - '%' comment field (any #) */
		"TITLE",                /* 11 - 'N' Title field           */
		"SURNAME",              /* 12 - 'n' Surname field         */
		"SIGNATURE1",			/* 13 - '_' User Sig field        */
		"SIGNATURE2",           /* 14 - '.' BBS Sig field         */
		"GROUP",                /* 15 - 'G'roup field             */
		"PHONE",                /* 16 - 'H' Phone number field    */
		"TWIT_REGION", 			/* 17 - 'J' Twit region field     */
		"TWIT_COUNTRY",         /* 18 - 'j' Twit country field    */
		"MAIL_TO_USER",         /* 19 - E-Mail 'T'o field         */
		"FORCED_PATH",          /* 20 - 'p' forced path field     */
		"REAL_NAME",            /* 21 - '!' Real name field       */
		"MAIL_TO_NODE",         /* 22 - 'Z' Destination node name */
		"MAIL_TO_ADD",          /* 23 - '>' Destination address   */
		"COUNTRY",				/* 24 - 'Q' county/state field    */
		"PUBLIC_TO",			/* 25 - 't' public 'to' field     */
		"FORMATTING",			/* 26 - if 1 then don't use ^A5   */
		"PART_SIZE",			/* 27 - the size of one message.  */
		"GATEWAY",				/* 28 - the '*' gateway field.    */
		"APPEND",				/* 29 - allow posting of multiple files. */
		"COMP",					/* 30 - send message to an extra file.. */
        "local",                // 31 - set the local-only bit.
        "maxsize",              // 32 - maxfsize setting.
        "minsize",              // 33 - minfsize.
        "accuracy",             // 34 - abbey_road.
        NULL
		};

	if ((cfgfile = fopen (cfg.configfile, "rt")) == NULL) {
		fnsplit(cfg.progname, drive, dir, NULL, NULL);
		sprintf(cfg.configfile, "%s%sposter.cfg", drive, dir);
		printf("Using config file: %s\n\n", cfg.configfile);
		if((cfgfile = fopen(cfg.configfile, "rt")) == 0) {
			printf ("No config file present.\n\n");
			return(FALSE);
		}
	}
	else
		printf("Config file \"%s\" found.\n\n", cfg.configfile);

	while (!feof(cfgfile))
	{
		text = fgets(line, sizeof(line), cfgfile);
		if(text == 0)
			break;
		if(line[0] != '#')
			continue;

		i = sscanf(line, "%s \"%255[^\"\n\r\t]", word1, lword1);

		if(i == 1)
			i = sscanf(line, "%s %255s", word1, lword1);
		if(i != 2)
			continue;

		for(i = 0; KEYWORD[i] != NULL; i++)
			if(stricmp(&word1[1], KEYWORD[i]) == 0)
			   break;

		switch (i)
		{
			case 0:
                lword1[LABELSIZE] = 0;
				strcpy(msgbuf.author, lword1);
                multi_author++;
				break;
			case 1:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.node, lword1);
				break;
			case 2:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.address, lword1);
				break;
			case 3:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.region, lword1);
				break;
			case 4:
				lword1[LABELSIZE] = 0;
				strcpy(cfg.room, lword1);
				break;
			case 5:
                strupr(lword1);
				if(lword1[0] == '1' || strstr(lword1, "YES") || strstr(lword1, "ON"))
					cfg.unix = TRUE;
				else
					cfg.unix = FALSE;
				break;
			case 6:
				cfg.startpoint = atol(lword1);
				break;
			case 7:
				lword1[LINESIZE] = 0;
				strcpy(cfg.startstr, lword1);
				break;
			case 8:
				lword1[BIGLABEL] = 0;
				strcpy(msgbuf.subject, lword1);
				break;
			case 9:
				lword1[LINESIZE] = 0;
				strcpy(cfg.endstr, lword1);
				break;
			case 10:
				lword1[LINESIZE] = 0;
                add_cmt(lword1);
				break;
			case 11:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.title, lword1);
				break;
			case 12:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.surname, lword1);
				break;
			case 13:
				lword1[BIGLABEL] = 0;
				strcpy(msgbuf.sig1, lword1);
				break;
			case 14:
				lword1[BIGLABEL] = 0;
				strcpy(msgbuf.sig2, lword1);
				break;
			case 15:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.group, lword1);
				break;
			case 16:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.phone, lword1);
				break;
			case 17:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.twit_region, lword1);
				break;
			case 18:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.twit_country, lword1);
				break;
			case 19:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.mail_to, lword1);
				break;
			case 20:
				lword1[LINESIZE] = 0;
				strcpy(msgbuf.fpath, lword1);
				break;
			case 21:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.realname, lword1);
				break;
			case 22:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.node_to, lword1);
				break;
			case 23:
				lword1[8] = 0;
				strcpy(msgbuf.address_to, lword1);
				break;
			case 24:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.state, lword1);
				break;
			case 25:
				lword1[LABELSIZE] = 0;
				strcpy(msgbuf.public_to, lword1);
				break;
			case 26:
                cfg.spaces = atoi(lword1);
				break;
			case 27:
				cfg.partsize = atoi(lword1);
				break;
			case 28:
				lword1[LINESIZE] = 0;
				strcpy(msgbuf.gateway, lword1);
				break;
			case 29:
                if(atoi(lword1) == 69)
                    cfg.append = TRUE;
                else
                    cfg.append = FALSE;
                break;
			case 30:
				lword1[LINESIZE] = 0;
				cfg.montyl = TRUE;
				strcpy(cfg.altname, lword1);
				break;

            case 31:
                strupr(lword1);
				if(lword1[0] == '1' || strstr(lword1, "YES")
                || strstr(lword1, "ON")) {
                    msgbuf.attr  = 0x40;
                    msgbuf.attr2 = 0x01;
                    cfg.local = TRUE;
                } else {
                    msgbuf.attr  = 0x00;
                    msgbuf.attr2 = 0x00;
					cfg.local = FALSE;
                }
				break;

            case 32:   // maxsize
                 cfg.maxfsize = atol(lword1);
                 break;

            case 33:   // minsize
                 cfg.minfsize = atol(lword1);
                 break;

            case 34:   // accuracy
                strupr(lword1);
				if(lword1[0] == '1' || strstr(lword1, "YES") || strstr(lword1, "ON"))
                    cfg.abbey_road = TRUE;
                else
                    cfg.abbey_road = FALSE;
                break;

			 default:
				printf("Keyword '%s' unknown.\n", word1);
				break;
		}
	}
	fclose(cfgfile);
	return(0);
}

/*************************************************************************/
/* envr() -- sets a couple of things from environmental vars...          */
/*************************************************************************/
void envr(void)
{
	char *ptr;

	if((ptr = getenv("SUBJECT")) != NULL) {
        *msgbuf.subject = 0;
		strncat(msgbuf.subject, ptr, BIGLABEL);
    }

	ptr = 0;
	if((ptr = getenv("GATEWAY")) != NULL) {
        *msgbuf.gateway = 0;
		strncat(msgbuf.gateway, ptr, LINESIZE);
    }
}

/****************************************************************************/
/* clean_msgbuf() - Cleans out the message buffer.  Calls clean_cmts().     */
/****************************************************************************/
void clean_msgbuf(void)
{
    msgbuf.ID            = 0;    // Local ID field.  Ignored by Citadel.
    *msgbuf.SID          = 0;    // Source ID
    *msgbuf.stamp        = 0;    // Time Stamp
    *msgbuf.author       = 0;
    *msgbuf.title        = 0;
    *msgbuf.surname      = 0;
    *msgbuf.group        = 0;
    *msgbuf.phone        = 0;
    *msgbuf.twit_region  = 0;
    *msgbuf.twit_country = 0;
    *msgbuf.software     = 0;
    *msgbuf.mail_to      = 0;     // Mail- recipient
    *msgbuf.node_to      = 0;     // Mail- node to send mail to
    *msgbuf.address_to   = 0;     // '>' Mail- Address of node to send mail to
    *msgbuf.public_to    = 0;     // 'f' Public- Public TO field.
    *msgbuf.realname     = 0;
    *msgbuf.node         = 0;
    *msgbuf.address      = 0;
    *msgbuf.room         = 0;
    *msgbuf.state        = 0;
    *msgbuf.region       = 0;
    *msgbuf.text         = 0;
    *msgbuf.subject      = 0;
    *msgbuf.sig1         = 0;
    *msgbuf.sig2         = 0;
    *msgbuf.fpath        = 0;
    *msgbuf.gateway      = 0;
    msgbuf.attr          = 0x00;
    msgbuf.attr2         = 0x00;
    clean_cmts();
    msgbuf.cmt_ptr       = NULL;
}

/****************************************************************************/
/* add_cmt() -- add a comment                                               */
/****************************************************************************/
int add_cmt(char *str)
{
comment *newcmt = NULL;

    if(msgbuf.cmt_ptr == NULL) {    // First comment.
        msgbuf.cmt_ptr = currcmt = (comment *)malloc(sizeof(comment));
        currcmt->next = NULL;
    } else {                      /* list has at least one entry. */
        if((newcmt = (comment *)malloc(sizeof(comment))) == NULL) {
            printf("Not enough room to store comment\n");
            return(FALSE);
        }
        newcmt->next = NULL;
        currcmt->next = newcmt;
        currcmt = newcmt;
    }

    strcpy(currcmt->text, str);
    numcmts++;
    return(TRUE);
}

/****************************************************************************/
/* remove_cmt() -- remove a comment.                                        */
/****************************************************************************/
int remove_cmt(void)
{
   if(msgbuf.cmt_ptr == NULL)        // No comments to remove...
       return(TRUE);

   if(msgbuf.cmt_ptr->next == NULL) { // Only one comment exists.
       free(msgbuf.cmt_ptr);
       msgbuf.cmt_ptr = NULL;
       numcmts = 0;
       return(TRUE);
   }
                                      // Several comments exist...
   for(currcmt = msgbuf.cmt_ptr; currcmt->next->next; currcmt = currcmt->next)
       ;       // Find the second to last comment.

   free(currcmt->next);
   currcmt->next = NULL;
   numcmts--;

   return(TRUE);
}

/****************************************************************************/
/* clean_cmts() -- remove all comments.                                     */
/****************************************************************************/
int     clean_cmts(void)
{
comment *cmt;

    currcmt = msgbuf.cmt_ptr;

    while(currcmt != NULL) {
        cmt = currcmt->next;
        free(currcmt);
        currcmt = cmt;
    }

    msgbuf.cmt_ptr = NULL;
    currcmt        = NULL;
    numcmts        = 0;

    return(TRUE);
}

/****************************************************************************/
/* reset() - Sets configuration defaults.                                   */
/****************************************************************************/
void reset(void)
{
    cfg.spaces        = 1;      // 0 = ^A5, 1 = ' ' (spaces)
    cfg.unix          = 0;      // 1 = Strip unix mail header.
    cfg.partsize      = 8000;   // Part size.  Default = 8000.
    cfg.append        = 0;      // allow appending to an existing packet.
    cfg.montyl        = 0;      // Output the message text to another file
    cfg.abbey_road    = 1;
    cfg.startpoint    = 0;      // # of lines to skip at start of file.
    cfg.minfsize      = 0;      // minimum file size to post.
    cfg.maxfsize      = 0;      // max file size to post.
    cfg.local         = 0;      // Set local-only bit?
    *cfg.room         = 0;      // Room name.
    *cfg.input        = 0;      // Input file + path.
    *cfg.altname      = 0;      // montyl - name of other file to print to.
    *cfg.endstr       = 0;      // The string to STOP posting on.
    *cfg.startstr     = 0;      // The string to START posting from.
    strcpy(cfg.configfile, "poster.cfg");
}

/****************************************************************************/
/* usage()                                                                  */
/****************************************************************************/
void usage(void)
{
    printf("More...");
    getch();                          // Pause.
    putc('\r', stdout);

	printf("Usage: %s [-c cfgfile] [-r room] [-a author] [-l+/-] [-f#] files...\n"
		   "\n"
           "       config file == name of config file.\n"
	       "              room == room messages show up in.\n"
	       "            author == author of messages.\n"
           "               -l+ == post local-only messages.\n"
           "               -l- == post network messages.\n"
           "               -f# == 0 = ^A5 formatting, 1 = spaces formatting,\n"
           "                      3 = intelligent formatting.\n"
           "              file == any number of filenames to post.\n"
           "\n"
           "Options can occure anywhere.  Options can appear more than once.\n"
           "Options only effect files to their _right_.\n"
           "\n"
           "eg: %s -r Foo file1 file2 -r \"err force\" file3 file4\n"
           "\n"
           "Files 1 and 2 will end up in Foo, 3 and 4 will be in err force.\n"
	       "\n"
           "Changing the config file resets all config options.  Only options set in the\n"
           "new config file will be in effect.  Config file changes are not additive.\n"
           "\n"
	       "Any multi-word args MUST be in quotes.\n", cfg.shrt_pname, cfg.shrt_pname);
}
